@@ -9,8 +9,9 @@ from configparser import MAX_INTERPOLATION_DEPTH, ConfigParser
 from configparser import Error as ConfigError
 from configparser import (
     ExtendedInterpolation, Interpolation, InterpolationDepthError,
-    InterpolationError, NoOptionError, NoSectionError,
+    InterpolationError, NoOptionError, NoSectionError, RawConfigParser,
 )
+from typing import Dict, List, Mapping, Optional, Sequence, Tuple, MutableMapping, Set
 
 import skytools
 
@@ -20,6 +21,24 @@ __all__ = (
 )
 
 
+def read_versioned_config(filenames: Sequence[str], main_section: str) -> ConfigParser:
+    """Pick syntax based on "config_format" value.
+    """
+    rcf = RawConfigParser()
+    rcf.read(filenames)
+
+    # avoid has_option here, so value can live in DEFAULT section
+    ver = rcf.get(main_section, "config_format", fallback="1")
+    if ver == "1":
+        cf = ConfigParser()
+    elif ver == "2":
+        cf = ExtendedConfigParser()
+    else:
+        raise ConfigError('Unsupported config format %r in %r' % (ver, filenames))
+    cf.read(filenames)
+    return cf
+
+
 class Config:
     """Bit improved ConfigParser.
 
@@ -28,8 +47,18 @@ class Config:
      - Accepts defaults in get() functions.
      - List value support.
     """
-    def __init__(self, main_section, filename, sane_config=None,
-                 user_defs=None, override=None, ignore_defs=False):
+    main_section: str               # main section
+    filename: Optional[str]         # file name that was loaded
+    override: Mapping[str, str]     # override values in config file
+    defs: Mapping[str, str]         # defaults visible in all sections
+    cf: ConfigParser                # actual ConfigParser instance
+
+    def __init__(self, main_section: str,
+                 filename: Optional[str],
+                 sane_config: Optional[bytes] = None,   # unused
+                 user_defs: Optional[Mapping[str, str]] = None,
+                 override: Optional[Mapping[str, str]] = None,
+                 ignore_defs: bool = False) -> None:
         """Initialize Config and read from file.
         """
         # use config file name as default job_name
@@ -56,16 +85,18 @@ class Config:
         self.main_section = main_section
         self.filename = filename
         self.override = override or {}
-        self.cf = ConfigParser()
 
         if filename is None:
+            self.cf = ConfigParser()
             self.cf.add_section(main_section)
         elif not os.path.isfile(filename):
             raise ConfigError('Config file not found: ' + filename)
+        else:
+            self.cf = read_versioned_config([filename], main_section)
 
         self.reload()
 
-    def reload(self):
+    def reload(self) -> None:
         """Re-reads config file."""
         if self.filename:
             self.cf.read(self.filename)
@@ -82,7 +113,7 @@ class Config:
             for k, v in self.override.items():
                 self.cf.set(self.main_section, k, v)
 
-    def get(self, key, default=None):
+    def get(self, key: str, default: Optional[str] = None) -> str:
         """Reads string value, if not set then default."""
 
         if not self.cf.has_option(self.main_section, key):
@@ -92,7 +123,7 @@ class Config:
 
         return str(self.cf.get(self.main_section, key))
 
-    def getint(self, key, default=None):
+    def getint(self, key: str, default: Optional[int] = None) -> int:
         """Reads int value, if not set then default."""
 
         if not self.cf.has_option(self.main_section, key):
@@ -102,7 +133,7 @@ class Config:
 
         return self.cf.getint(self.main_section, key)
 
-    def getboolean(self, key, default=None):
+    def getboolean(self, key: str, default: Optional[bool] = None) -> bool:
         """Reads boolean value, if not set then default."""
 
         if not self.cf.has_option(self.main_section, key):
@@ -112,7 +143,7 @@ class Config:
 
         return self.cf.getboolean(self.main_section, key)
 
-    def getfloat(self, key, default=None):
+    def getfloat(self, key: str, default: Optional[float] = None) -> float:
         """Reads float value, if not set then default."""
 
         if not self.cf.has_option(self.main_section, key):
@@ -122,7 +153,7 @@ class Config:
 
         return self.cf.getfloat(self.main_section, key)
 
-    def getlist(self, key, default=None):
+    def getlist(self, key: str, default: Optional[Sequence[str]] = None) -> Sequence[str]:
         """Reads comma-separated list from key."""
 
         if not self.cf.has_option(self.main_section, key):
@@ -131,14 +162,14 @@ class Config:
             return default
 
         s = self.get(key).strip()
-        res = []
+        res: List[str] = []
         if not s:
             return res
         for v in s.split(","):
             res.append(v.strip())
         return res
 
-    def getdict(self, key, default=None):
+    def getdict(self, key: str, default: Optional[Mapping[str, str]] = None) -> Mapping[str, str]:
         """Reads key-value dict from parameter.
 
         Key and value are separated with ':'.  If missing,
@@ -151,7 +182,7 @@ class Config:
             return default
 
         s = self.get(key).strip()
-        res = {}
+        res: Dict[str, str] = {}
         if not s:
             return res
         for kv in s.split(","):
@@ -165,7 +196,7 @@ class Config:
             res[k] = v
         return res
 
-    def getfile(self, key, default=None):
+    def getfile(self, key: str, default: Optional[str] = None) -> str:
         """Reads filename from config.
 
         In addition to reading string value, expands ~ to user directory.
@@ -181,7 +212,7 @@ class Config:
 
         return fn
 
-    def getbytes(self, key, default=None):
+    def getbytes(self, key: str, default: Optional[str] = None) -> int:
         """Reads a size value in human format, if not set then default.
 
         Examples: 1, 2 B, 3K, 4 MB
@@ -196,7 +227,7 @@ class Config:
 
         return skytools.hsize_to_bytes(s)
 
-    def get_wildcard(self, key, values=(), default=None):
+    def get_wildcard(self, key: str, values: Sequence[str] = (), default: Optional[str] = None) -> str:
         """Reads a wildcard property from conf and returns its string value, if not set then default."""
 
         orig_key = key
@@ -215,27 +246,27 @@ class Config:
             raise NoOptionError(orig_key, self.main_section)
         return default
 
-    def sections(self):
+    def sections(self) -> Sequence[str]:
         """Returns list of sections in config file, excluding DEFAULT."""
         return self.cf.sections()
 
-    def has_section(self, section):
+    def has_section(self, section: str) -> bool:
         """Checks if section is present in config file, excluding DEFAULT."""
         return self.cf.has_section(section)
 
-    def clone(self, main_section):
+    def clone(self, main_section: str) -> "Config":
         """Return new Config() instance with new main section on same config file."""
         return Config(main_section, self.filename)
 
-    def options(self):
+    def options(self) -> Sequence[str]:
         """Return list of options in main section."""
         return self.cf.options(self.main_section)
 
-    def has_option(self, opt):
+    def has_option(self, opt: str) -> bool:
         """Checks if option exists in main section."""
         return self.cf.has_option(self.main_section, opt)
 
-    def items(self):
+    def items(self) -> Sequence[Tuple[str, str]]:
         """Returns list of (name, value) for each option in main section."""
         return self.cf.items(self.main_section)
 
@@ -243,24 +274,29 @@ class Config:
     getbool = getboolean
 
 
+ParserSection = Mapping[str, str]
+ParserState = MutableMapping[str, ParserSection]
+#ParserState = ConfigParser
+ParserLoop = Set[Tuple[str, str]]
+
 class ExtendedInterpolationCompat(Interpolation):
     _EXT_VAR_RX = r'\$\$|\$\{[^(){}]+\}'
     _OLD_VAR_RX = r'%%|%\([^(){}]+\)s'
     _var_rc = re.compile('(%s|%s)' % (_EXT_VAR_RX, _OLD_VAR_RX))
     _bad_rc = re.compile('[%$]')
 
-    def before_get(self, parser, section, option, rawval, defaults):
-        dst = []
-        self._interpolate_ext(dst, parser, section, option, rawval, defaults, set())
+    def before_get(self, parser: ParserState, section: str, option: str, value: str, defaults: ParserSection) -> str:
+        dst: List[str] = []
+        self._interpolate_ext(dst, parser, section, option, value, defaults, set())
         return ''.join(dst)
 
-    def before_set(self, parser, section, option, value):
+    def before_set(self, parser: ParserState, section: str, option: str, value: str) -> str:
         sub = self._var_rc.sub('', value)
         if self._bad_rc.search(sub):
             raise ValueError("invalid interpolation syntax in %r" % value)
         return value
 
-    def _interpolate_ext(self, dst, parser, section, option, rawval, defaults, loop_detect):
+    def _interpolate_ext(self, dst: List[str], parser: ParserState, section: str, option: str, rawval: str, defaults: ParserSection, loop_detect: ParserLoop) -> None:
         if not rawval:
             return
 
@@ -275,7 +311,7 @@ class ExtendedInterpolationCompat(Interpolation):
         parts = self._var_rc.split(rawval)
         for i, frag in enumerate(parts):
             fullkey = None
-            use_vars = defaults
+            use_vars: Optional[ParserSection] = defaults
             if i % 2 == 0:
                 dst.append(frag)
                 continue
@@ -297,11 +333,12 @@ class ExtendedInterpolationCompat(Interpolation):
             else:
                 raise InterpolationError(section, option, 'Internal parse error: %r' % frag)
 
-            key = parser.optionxform(key)
-            newpart = parser.get(ksect, key, raw=True, vars=use_vars)
-            if newpart is None:
-                raise InterpolationError(ksect, key, 'Key referenced is None')
-            self._interpolate_ext(dst, parser, ksect, key, newpart, defaults, loop_detect)
+            if isinstance(parser, RawConfigParser):
+                key = parser.optionxform(key)
+                newpart = parser.get(ksect, key, raw=True, vars=use_vars)
+                if newpart is None:
+                    raise InterpolationError(ksect, key, 'Key referenced is None')
+                self._interpolate_ext(dst, parser, ksect, key, newpart, defaults, loop_detect)
 
         loop_detect.remove(xloop)
 
@@ -311,7 +348,7 @@ class ExtendedConfigParser(ConfigParser):
 
     Syntax: ${var} and ${section:var}
     """
-    _DEFAULT_INTERPOLATION = ExtendedInterpolation()
+    _DEFAULT_INTERPOLATION: Interpolation = ExtendedInterpolation()
 
 
 class ExtendedCompatConfigParser(ExtendedConfigParser):
@@ -320,5 +357,5 @@ class ExtendedCompatConfigParser(ExtendedConfigParser):
     New ${} syntax allows ${key} to refer key in same section,
     and ${sect:key} to refer key in other sections.
     """
-    _DEFAULT_INTERPOLATION = ExtendedInterpolationCompat()
+    _DEFAULT_INTERPOLATION: Interpolation = ExtendedInterpolationCompat()
 
