@@ -58,7 +58,7 @@ def render_table(rows, fields):
     lines.append(fmt % tuple(fields))
     lines.append(fmt % tuple(['-' * 15] * len(fields)))
     for row in rows:
-        lines.append(fmt % tuple([str(row.get(k)) for k in fields]))
+        lines.append(fmt % tuple(str(row.get(k)) for k in fields))
     return lines
 
 
@@ -475,7 +475,7 @@ class ServiceContext(DBService):
     def __init__(self, context, global_dict=None):
         """ This object must be initiated in the beginning of each db service
         """
-        super(ServiceContext, self).__init__(context, global_dict)
+        super().__init__(context, global_dict)
 
         rec = skytools.db_urldecode(context)
         if "username" not in rec:
