@@ -1,7 +1,10 @@
 """Find table structure and allow CREATE/DROP elements from it.
 """
 
+# pylint:disable=arguments-renamed
+
 import re
+from typing import List
 
 import skytools
 from skytools import quote_fqident, quote_ident
@@ -576,7 +579,7 @@ class BaseStruct:
     Allow to issue CREATE/DROP statements about any
     group of elements.
     """
-    object_list = []
+    object_list: List[TElem] = []
     def __init__(self, curs, name):
         """Initializes class by loading info about table_name from database."""
 
@@ -645,7 +648,7 @@ class TableStruct(BaseStruct):
     def __init__(self, curs, table_name):
         """Initializes class by loading info about table_name from database."""
 
-        super(TableStruct, self).__init__(curs, table_name)
+        super().__init__(curs, table_name)
 
         self.table_name = table_name
 
@@ -707,7 +710,7 @@ class SeqStruct(BaseStruct):
     def __init__(self, curs, seq_name):
         """Initializes class by loading info about table_name from database."""
 
-        super(SeqStruct, self).__init__(curs, seq_name)
+        super().__init__(curs, seq_name)
 
         # fill args
         args = {
