@@ -1,25 +1,35 @@
 """File utilities
 """
 
+# pylint:disable=unspecified-encoding
+
 import errno
 import os
 import sys
+from typing import Optional, Union
 
 __all__ = ['write_atomic', 'signal_pidfile']
 
 
-def write_atomic_unix(fn, data, bakext=None, mode='b'):
+def write_atomic_unix(fn: str, data: Union[bytes, str], bakext: Optional[str] = None, mode: str = 'b'):
     """Write file with rename."""
 
     if mode not in ['', 'b', 't']:
         raise ValueError("unsupported fopen mode")
-    if mode == 'b' and not isinstance(data, bytes):
+
+    if mode == "b" and not isinstance(data, bytes):
         data = data.encode('utf8')
 
+    mode = 'w' + mode
+
     # write new data to tmp file
     fn2 = fn + '.new'
-    with open(fn2, 'w' + mode) as f:
-        f.write(data)
+    if "b" in mode:
+        with open(fn2, mode) as f:
+            f.write(data)
+    else:
+        with open(fn2, mode, encoding="utf8") as f:
+            f.write(data)
 
     # link old data to bak file
     if bakext:
@@ -48,7 +58,7 @@ def write_atomic_unix(fn, data, bakext=None, mode='b'):
     os.rename(fn2, fn)
 
 
-def signal_pidfile(pidfile, sig):
+def signal_pidfile(pidfile: str, sig: int) -> bool:
     """Send a signal to process whose ID is located in pidfile.
 
     Read only first line of pidfile to support multiline
@@ -60,7 +70,7 @@ def signal_pidfile(pidfile, sig):
 
     ln = ''
     try:
-        with open(pidfile, 'r') as f:
+        with open(pidfile, 'r', encoding="utf8") as f:
             ln = f.readline().strip()
         pid = int(ln)
         if sig == 0 and sys.platform == 'win32':
@@ -75,11 +85,11 @@ def signal_pidfile(pidfile, sig):
         # but more common case is old empty file.
         if not ln:
             return False
-        raise ValueError('Corrupt pidfile: %s' % pidfile)
+        raise ValueError('Corrupt pidfile: %s' % pidfile) from None
     return False
 
 
-def win32_detect_pid(pid):
+def win32_detect_pid(pid: int) -> bool:
     """Process detection for win32."""
 
     # avoid pywin32 dependecy, use ctypes instead
@@ -92,7 +102,7 @@ def win32_detect_pid(pid):
     ERROR_ACCESS_DENIED = 5
 
     # Load kernel32.dll
-    k = ctypes.windll.kernel32
+    k = ctypes.windll.kernel32  # type: ignore
     OpenProcess = k.OpenProcess
     OpenProcess.restype = ctypes.c_void_p
 
@@ -111,19 +121,25 @@ def win32_detect_pid(pid):
     return code.value == STILL_ACTIVE
 
 
-def win32_write_atomic(fn, data, bakext=None, mode='b'):
+def win32_write_atomic(fn: str, data: Union[bytes, str], bakext: Optional[str] = None, mode: str = 'b'):
     """Write file with rename for win32."""
 
     if mode not in ['', 'b', 't']:
         raise ValueError("unsupported fopen mode")
 
-    if mode == 'b' and not isinstance(data, bytes):
+    if mode == "b" and not isinstance(data, bytes):
         data = data.encode('utf8')
 
+    mode = "w" + mode
+
     # write new data to tmp file
     fn2 = fn + '.new'
-    with open(fn2, 'w' + mode) as f:
-        f.write(data)
+    if "b" in mode:
+        with open(fn2, mode) as f:
+            f.write(data)
+    else:
+        with open(fn2, mode, encoding="utf8") as f:
+            f.write(data)
 
     # move old data to bak file
     if bakext:
