@@ -2,6 +2,7 @@
 """
 
 import re
+from typing import Iterator, List, Optional, Sequence, Tuple, Dict, Union
 
 import skytools
 
@@ -15,7 +16,7 @@ __all__ = (
 _rc_listelem = re.compile(r'( [^,"}]+ | ["] ( [^"\\]+ | [\\]. )* ["] )', re.X)
 
 
-def parse_pgarray(array):
+def parse_pgarray(array: Optional[str]) -> Optional[List[Optional[str]]]:
     r"""Parse Postgres array and return list of items inside it.
     """
     if array is None:
@@ -59,11 +60,11 @@ def parse_pgarray(array):
 
 class _logtriga_parser:
     """Parses logtriga/sqltriga partial SQL to values."""
-    pklist = None
-    def tokenizer(self, sql):
+    pklist: Optional[Sequence[str]] = None
+    def tokenizer(self, sql: str) -> Iterator[str]:
         """Token generator."""
-        for ___typ, tok in sql_tokenizer(sql, ignore_whitespace=True):
-            yield tok
+        for tup in sql_tokenizer(sql, ignore_whitespace=True):
+            yield tup[1]
 
     def parse_insert(self, tk, fields, values, key_fields, key_values):
         """Handler for inserts."""
@@ -137,17 +138,18 @@ class _logtriga_parser:
         values = [skytools.unquote_literal(f) for f in values]
         return skytools.dbdict(zip(fields, values))
 
-    def parse_sql(self, op, sql, pklist=None, splitkeys=False):
+    def parse_sql(self, op: str, sql: str, pklist: Optional[Sequence[str]] = None, splitkeys: bool = False
+    ) -> Union[skytools.dbdict, Tuple[skytools.dbdict, skytools.dbdict]]:
         """Main entry point."""
         if pklist is None:
             self.pklist = []
         else:
             self.pklist = pklist
         tk = self.tokenizer(sql)
-        fields = []
-        values = []
-        key_fields = []
-        key_values = []
+        fields: List[str] = []
+        values: List[str] = []
+        key_fields: List[str] = []
+        key_values: List[str] = []
         try:
             if op == "I":
                 self.parse_insert(tk, fields, values, key_fields, key_values)
@@ -169,11 +171,15 @@ class _logtriga_parser:
         return self._create_dbdict(fields + key_fields, values + key_values)
 
 
-def parse_logtriga_sql(op, sql, splitkeys=False):
+def parse_logtriga_sql(
+    op: str, sql: str, splitkeys: bool = False
+) -> Union[skytools.dbdict, Tuple[skytools.dbdict, skytools.dbdict]]:
     return parse_sqltriga_sql(op, sql, splitkeys=splitkeys)
 
 
-def parse_sqltriga_sql(op, sql, pklist=None, splitkeys=False):
+def parse_sqltriga_sql(
+    op: str, sql: str, pklist: Optional[Sequence[str]] = None, splitkeys: bool = False
+) -> Union[skytools.dbdict, Tuple[skytools.dbdict, skytools.dbdict]]:
     """Parse partial SQL used by pgq.sqltriga() back to data values.
 
     Parser has following limitations:
@@ -188,7 +194,7 @@ def parse_sqltriga_sql(op, sql, pklist=None, splitkeys=False):
     return _logtriga_parser().parse_sql(op, sql, pklist, splitkeys=splitkeys)
 
 
-def parse_tabbed_table(txt):
+def parse_tabbed_table(txt: str) -> List[Dict[str, str]]:
     r"""Parse a tab-separated table into list of dicts.
 
     Expect first row to be column names.
@@ -198,7 +204,7 @@ def parse_tabbed_table(txt):
 
     txt = txt.replace("\r\n", "\n")
     fields = None
-    data = []
+    data: List[Dict[str, str]] = []
     for ln in txt.split("\n"):
         if not ln:
             continue
@@ -243,8 +249,10 @@ _std_sql_rc = _ext_sql_rc = None
 _std_sql_fq_rc = _ext_sql_fq_rc = None
 
 
-def sql_tokenizer(sql, standard_quoting=False, ignore_whitespace=False,
-                  fqident=False, show_location=False):
+def sql_tokenizer(
+        sql: str, standard_quoting: bool = False, ignore_whitespace: bool = False,
+        fqident: bool = False, show_location: bool = False
+        ) -> Iterator[Union[Tuple[str, str], Tuple[str, str, int]]]:
     r"""Parser SQL to tokens.
 
     Iterator, returns (toktype, tokstr) tuples.
@@ -273,7 +281,7 @@ def sql_tokenizer(sql, standard_quoting=False, ignore_whitespace=False,
         if not m:
             break
         pos = m.end()
-        typ = m.lastgroup
+        typ = m.lastgroup or '?'
         if ignore_whitespace and typ == "ws":
             continue
         tk = m.group()
@@ -287,7 +295,7 @@ _copy_from_stdin_re = r"copy.*from\s+stdin"
 _copy_from_stdin_rc = None
 
 
-def parse_statements(sql, standard_quoting=False):
+def parse_statements(sql: str, standard_quoting: bool = False) -> Iterator[str]:
     """Parse multi-statement string into separate statements.
 
     Returns list of statements.
@@ -296,9 +304,10 @@ def parse_statements(sql, standard_quoting=False):
     global _copy_from_stdin_rc
     if not _copy_from_stdin_rc:
         _copy_from_stdin_rc = re.compile(_copy_from_stdin_re, re.X | re.I)
-    tokens = []
+    tokens: List[str] = []
     pcount = 0  # '(' level
-    for typ, t in sql_tokenizer(sql, standard_quoting=standard_quoting):
+    for tmp in sql_tokenizer(sql, standard_quoting=standard_quoting):
+        typ, t = tmp[0], tmp[1]
         # skip whitespace and comments before statement
         if len(tokens) == 0 and typ == "ws":
             continue
@@ -331,7 +340,7 @@ _acl_re = r'''
 _acl_rc = None
 
 
-def parse_acl(acl):
+def parse_acl(acl: str) -> Optional[Tuple[str, str, str]]:
     """Parse ACL entry.
     """
     global _acl_rc
@@ -356,7 +365,7 @@ def parse_acl(acl):
     return (target, perm, owner)
 
 
-def dedent(doc):
+def dedent(doc: str) -> str:
     r"""Relaxed dedent.
 
     - takes whitespace to be removed from first indented line.
@@ -366,8 +375,8 @@ def dedent(doc):
     - ignores indent of empty lines
     - if line does not match common indent, is stays unchanged
     """
-    pfx = None
-    res = []
+    pfx: Optional[str] = None
+    res: List[str] = []
     for ln in doc.splitlines():
         ln = ln.rstrip()
         if not pfx and len(res) < 2:
@@ -383,7 +392,7 @@ def dedent(doc):
     return '\n'.join(res)
 
 
-def hsize_to_bytes(input_str):
+def hsize_to_bytes(input_str: str) -> int:
     """ Convert sizes from human format to bytes (string to integer)
     """
 
@@ -407,7 +416,7 @@ _cstr_unesc_rc = None
 _cstr_badval_rc = None
 
 
-def parse_connect_string(cstr):
+def parse_connect_string(cstr: str) -> List[Tuple[str, str]]:
     r"""Parse Postgres connect string.
     """
     global _cstr_rc, _cstr_unesc_rc
@@ -429,7 +438,7 @@ def parse_connect_string(cstr):
     return res
 
 
-def merge_connect_string(cstr_arg_list):
+def merge_connect_string(cstr_arg_list: Sequence[Tuple[str, str]]) -> str:
     """Put fragments back together.
     """
     global _cstr_badval_rc
