@@ -9,8 +9,15 @@ If C{type} is missing, C{text} is assumed.
 See L{plpy_exec} for examples.
 """
 
+import json
+import re
+from functools import lru_cache
+from typing import Any, Dict, List, Mapping, Optional, Sequence, Union
+
 import skytools
 
+from .basetypes import Cursor
+
 try:
     import plpy
 except ImportError:
@@ -25,22 +32,44 @@ PARAM_INLINE = 0  # quote_literal()
 PARAM_DBAPI = 1  # %()s
 PARAM_PLPY = 2   # $n
 
+_RC_PARAM = re.compile(r"""
+    \{  ( [^|{}:]* )
+        (?:  : ( [^|{}:]+ ) )?
+        (?: \| ( [^|{}:]+ ) )?
+    ( \} )?
+""", re.X)
+
+
+def _inline_to_text(val: Any) -> Optional[str]:
+    """Approx emulate PL/Python and Psycopg2 internal conversions
+    for common types.
+    """
+    if val is None or isinstance(val, str):
+        return val
+    if isinstance(val, dict):
+        return json.dumps(val)
+    if isinstance(val, (tuple, list)):
+        return skytools.make_pgarray(val)
+    if isinstance(val, bytes):
+        return "\\x" + val.hex()
+    return str(val)
+
 
 class QArgConf:
     """Per-query arg-type config object."""
-    param_type = None
+    param_type = PARAM_INLINE
 
 
 class QArg:
     """Place-holder for a query parameter."""
-    def __init__(self, name, value, pos, conf):
+    def __init__(self, name: str, value: Any, pos: int, conf: QArgConf):
         self.name = name
         self.value = value
         self.pos = pos
         self.conf = conf
-    def __str__(self):
+    def __str__(self) -> str:
         if self.conf.param_type == PARAM_INLINE:
-            return skytools.quote_literal(self.value)
+            return skytools.quote_literal(_inline_to_text(self.value))
         elif self.conf.param_type == PARAM_DBAPI:
             return "%s"
         elif self.conf.param_type == PARAM_PLPY:
@@ -49,88 +78,37 @@ class QArg:
             raise Exception("bad QArgConf.param_type")
 
 
-# need an structure with fast remove-from-middle
-# and append operations.
-class DList:
-    """Simple double-linked list."""
-    __slots__ = ('next', 'prev')
-    def __init__(self):
-        self.next = self
-        self.prev = self
-
-    def append(self, obj):
-        obj.next = self
-        obj.prev = self.prev
-        self.prev.next = obj
-        self.prev = obj
-
-    def remove(self, obj):
-        obj.next.prev = obj.prev
-        obj.prev.next = obj.next
-        obj.next = obj.prev = None
-
-    def empty(self):
-        return self.next is self
-
-    def pop(self):
-        """Remove and return first element."""
-        obj = None
-        if not self.empty():
-            obj = self.next
-            self.remove(obj)
-        return obj
-
-
-class CachedPlan(DList):
-    """Wrapper around prepared plan."""
-    __slots__ = ('key', 'plan')
-    def __init__(self, key, plan):
-        super(CachedPlan, self).__init__()
-        self.key = key  # (sql, (types))
-        self.plan = plan
-
-
 class PlanCache:
     """Cache for limited amount of plans."""
 
-    def __init__(self, maxplans=100):
+    def __init__(self, maxplans: int = 100):
         self.maxplans = maxplans
-        self.plan_map = {}
-        self.plan_list = DList()
 
-    def get_plan(self, sql, types):
+        @lru_cache(maxplans)
+        def _cached_prepare(key):
+            sql, types = key
+            return plpy.prepare(sql, types)
+
+        self._cached_prepare = _cached_prepare
+
+    def get_plan(self, sql: str, types: Sequence[str]):
         """Prepare the plan and cache it."""
-
-        t = (sql, tuple(types))
-        if t in self.plan_map:
-            pc = self.plan_map[t]
-            # put to the end
-            self.plan_list.remove(pc)
-            self.plan_list.append(pc)
-            return pc.plan
-
-        # prepare new plan
-        plan = plpy.prepare(sql, types)
-
-        # add to cache
-        pc = CachedPlan(t, plan)
-        self.plan_list.append(pc)
-        self.plan_map[t] = pc
-
-        # remove plans if too much
-        while len(self.plan_map) > self.maxplans:
-            # this is ugly workaround for pylint
-            drop = self.plan_list.pop()
-            del self.plan_map[getattr(drop, 'key')]
-
-        return plan
+        key = (sql, tuple(types))
+        return self._cached_prepare(key)
 
 
 class QueryBuilderCore:
     """Helper for query building.
     """
 
-    def __init__(self, sqlexpr, params):
+    _params: Optional[Mapping[str, Any]]
+    _arg_type_list: List[str]
+    _arg_value_list: List[Any]
+    _sql_parts: List[Union[str, QArg]]
+    _arg_conf: QArgConf
+    _nargs: int
+
+    def __init__(self, sqlexpr: str, params: Optional[Mapping[str, Any]]):
         """Init the object.
 
         @param sqlexpr:     Partial sql fragment.
@@ -146,12 +124,12 @@ class QueryBuilderCore:
         if sqlexpr:
             self.add(sqlexpr, required=True)
 
-    def add(self, expr, sql_type="text", required=False):
+    def add(self, expr: str, sql_type: str = "text", required: bool = False):
         """Add SQL fragment to query.
         """
         self._add_expr('', expr, self._params, sql_type, required)
 
-    def get_sql(self, param_type=PARAM_INLINE):
+    def get_sql(self, param_type: int = PARAM_INLINE):
         """Return generated SQL (thus far) as string.
 
         Possible values for param_type:
@@ -163,45 +141,41 @@ class QueryBuilderCore:
         tmp = [str(part) for part in self._sql_parts]
         return "".join(tmp)
 
-    def _add_expr(self, pfx, expr, params, sql_type, required):
-        parts = []
-        types = []
-        values = []
+    def _add_expr(self, pfx: str, expr: str, params: Optional[Mapping[str, Any]], sql_type: str, required: bool):
+        parts: List[Union[str, QArg]] = []
+        types: List[str] = []
+        values: List[Any] = []
         nargs = self._nargs
         if pfx:
             parts.append(pfx)
         pos = 0
         while True:
-            # find start of next argument
-            a1 = expr.find('{', pos)
-            if a1 < 0:
+            # find next argument
+            m = _RC_PARAM.search(expr, pos)
+            if not m:
                 parts.append(expr[pos:])
                 break
 
-            # find end end of argument name
-            a2 = expr.find('}', a1)
-            if a2 < 0:
-                raise Exception("missing argument terminator: " + expr)
-
             # add plain sql
-            if a1 > pos:
-                parts.append(expr[pos:a1])
-            pos = a2 + 1
+            parts.append(expr[pos:m.start()])
+            pos = m.end()
 
-            # get arg name, check if exists
-            k = expr[a1 + 1: a2]
-            # split name from type
-            tpos = k.rfind(':')
-            if tpos > 0:
-                kparam = k[:tpos]
-                ktype = k[tpos + 1:]
-            else:
-                kparam = k
+            # get arg name and type
+            kparam, ktype, alt_frag, tag = m.groups()
+            if not kparam or not tag:
+                raise ValueError("invalid tag syntax: <%s>" % m.group(0))
+            if not ktype:
                 ktype = sql_type
 
             # params==None means params are checked later
-            if params is not None and kparam not in params:
-                if required:
+            if params is None:
+                if alt_frag is not None:
+                    raise ValueError("alt_frag not supported with params=None")
+            elif kparam not in params:
+                if alt_frag is not None:
+                    parts.append(alt_frag)
+                    continue
+                elif required:
                     raise Exception("required parameter missing: " + kparam)
                 # optional fragment, param missing, skip it
                 return
@@ -227,23 +201,24 @@ class QueryBuilderCore:
 
 
 class QueryBuilder(QueryBuilderCore):
-    def execute(self, curs):
+    def execute(self, curs: Cursor) -> None:
         """Client-side query execution on DB-API 2.0 cursor.
 
         Calls C{curs.execute()} with proper arguments.
 
-        Returns result of curs.execute(), although that does not
-        return anything interesting.  Later curs.fetch* methods
+        Does not return anything, curs.fetch* methods
         must be called to get result.
         """
         q = self.get_sql(PARAM_DBAPI)
         args = self._params
-        return curs.execute(q, args)
+        curs.execute(q, args)
 
 
 class PLPyQueryBuilder(QueryBuilderCore):
 
-    def __init__(self, sqlexpr, params, plan_cache=None, sqls=None):
+    def __init__(self, sqlexpr: str, params: Optional[Mapping[str, Any]],
+                 plan_cache: Optional[Dict[str, Any]] = None,
+                 sqls: Optional[List[Dict[str, str]]] = None):
         """Init the object.
 
         @param sqlexpr:     Partial sql fragment.
@@ -253,7 +228,7 @@ class PLPyQueryBuilder(QueryBuilderCore):
                             to query.  Usually either C{GD} or C{SD} should be given here.
         @param sqls:        list object where to append executed sqls (used for debugging)
         """
-        super(PLPyQueryBuilder, self).__init__(sqlexpr, params)
+        super().__init__(sqlexpr, params)
         self._sqls = sqls
 
         if plan_cache is not None:
@@ -263,7 +238,7 @@ class PLPyQueryBuilder(QueryBuilderCore):
         else:
             self._plan_cache = None
 
-    def execute(self):
+    def execute(self) -> Sequence[skytools.dbdict]:
         """Server-side query execution via plpy.
 
         Query can be run either cached or uncached, depending
@@ -278,13 +253,13 @@ class PLPyQueryBuilder(QueryBuilderCore):
         if self._sqls is not None:
             self._sqls.append({"sql": self.get_sql(PARAM_INLINE)})
 
+        sql = self.get_sql(PARAM_PLPY)
         if self._plan_cache is not None:
-            sql = self.get_sql(PARAM_PLPY)
             plan = self._plan_cache.get_plan(sql, types)
-            res = plpy.execute(plan, args)
         else:
-            sql = self.get_sql(PARAM_INLINE)
-            res = plpy.execute(sql)
+            plan = plpy.prepare(sql, types)
+
+        res = plpy.execute(plan, args)
         if res:
             res = [skytools.dbdict(r) for r in res]
         return res
@@ -295,7 +270,7 @@ class PLPyQuery:
 
     See L{plpy_exec} for simple usage.
     """
-    def __init__(self, sql):
+    def __init__(self, sql: str):
         qb = QueryBuilder(sql, None)
         p_sql = qb.get_sql(PARAM_PLPY)
         p_types = qb._arg_type_list
@@ -303,25 +278,32 @@ class PLPyQuery:
         self.arg_map = qb._arg_value_list
         self.sql = sql
 
-    def execute(self, arg_dict, all_keys_required=True):
+    def execute(self, arg_dict: Mapping[str, Any], all_keys_required=True) -> Sequence[skytools.dbdict]:
         try:
             if all_keys_required:
                 arg_list = [arg_dict[k] for k in self.arg_map]
             else:
                 arg_list = [arg_dict.get(k) for k in self.arg_map]
-            return plpy.execute(self.plan, arg_list)
+            res = plpy.execute(self.plan, arg_list)
+            if res:
+                return [skytools.dbdict(row) for row in res]
+            return res
         except KeyError:
             need = set(self.arg_map)
             got = set(arg_dict.keys())
             missing = list(need.difference(got))
             plpy.error("Missing arguments: [%s]  QUERY: %s" % (
                 ','.join(missing), repr(self.sql)))
+            raise ValueError("unreachable") from None
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return 'PLPyQuery<%s>' % self.sql
 
 
-def plpy_exec(gd, sql, args, all_keys_required=True):
+def plpy_exec(gd: Optional[Dict[str, Any]],
+              sql: str,
+              args: Optional[Mapping[str, Any]],
+              all_keys_required=True) -> Sequence[skytools.dbdict]:
     """Cached plan execution for PL/Python.
 
     @param gd:  dict to store cached plans under.  If None, caching is disabled.
@@ -345,7 +327,10 @@ def plpy_exec(gd, sql, args, all_keys_required=True):
 
 # some helper functions for convenient sql execution
 
-def run_query(cur, sql, params=None, **kwargs):
+def run_query(cur: Cursor, sql: str,
+              params: Optional[Mapping[str, Any]] = None,
+              **kwargs: Any
+              ) -> List[skytools.dbdict]:
     """ Helper function if everything you need is just paramertisized execute
         Sets rows_found that is coneninet to use when you don't need result just
         want to know how many rows were affected
@@ -356,11 +341,12 @@ def run_query(cur, sql, params=None, **kwargs):
     rows = cur.fetchall()
     # convert result rows to dbdict
     if rows:
-        rows = [skytools.dbdict(r) for r in rows]
-    return rows
+        return [skytools.dbdict(r) for r in rows]
+    return []
 
 
-def run_query_row(cur, sql, params=None, **kwargs):
+def run_query_row(cur: Cursor, sql: str, params: Optional[Mapping[str, Any]] = None, **kwargs: Any
+                  ) -> Optional[skytools.dbdict]:
     """ Helper function if everything you need is just paramertisized execute to
         fetch one row only. If not found none is returned
     """
@@ -371,7 +357,7 @@ def run_query_row(cur, sql, params=None, **kwargs):
     return rows[0]
 
 
-def run_lookup(cur, sql, params=None, **kwargs):
+def run_lookup(cur: Cursor, sql: str, params: Optional[Mapping[str, Any]] = None, **kwargs: Any) -> Any:
     """ Helper function to fetch one value Takes away all the hassle of preparing statements
         and processing returned result giving out just one value.
     """
@@ -384,7 +370,7 @@ def run_lookup(cur, sql, params=None, **kwargs):
     return row[0]
 
 
-def run_exists(cur, sql, params=None, **kwargs):
+def run_exists(cur: Cursor, sql: str, params: Optional[Mapping[str, Any]] = None, **kwargs: Any) -> bool:
     """ Helper function to fetch one value Takes away all the hassle of preparing statements
         and processing returned result giving out just one value.
     """
@@ -395,7 +381,7 @@ def run_exists(cur, sql, params=None, **kwargs):
 
 # fake plpy for testing
 class fake_plpy:
-    log = []
+    log: List[str] = []
     def prepare(self, sql, types):
         self.log.append("DBG: plpy.prepare(%s, %s)" % (repr(sql), repr(types)))
         return ('PLAN', sql, types)
@@ -405,10 +391,11 @@ class fake_plpy:
 
     def error(self, msg):
         self.log.append("DBG: plpy.error(%s)" % repr(msg))
+        raise Exception("plpy.error")
 
 
 # make plpy available
 if not plpy:
     plpy = fake_plpy()
-    GD = {}
+    GD: Dict[str, Any] = {}
 
